/*
 * FMV.C - File Majority voter
 * Create output file, where each byte in output file contains the byte value
 * found in the majority of the files input
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

FILE *image_handle;
int *count_array;
char **memptr_array;
char *buf_array;
FILE **fh_array;
int infile_count;
const int memblocksize = 1000000;

void cleanup(void)
{
	int i;

	if (fh_array)
		for (i = 0; i < infile_count; i++)
			if (fh_array[i])
				fclose(fh_array[i]);

	if (image_handle)
		fclose(image_handle);

	if (buf_array)
		free(buf_array);

	if (count_array)
		free(count_array);

	if (memptr_array)
		free(memptr_array);

}

void main(int argc, char *argv[])
{
	int i, j, k, min_count_max;

	if (argc < 3) {
		printf("Usage: %s <outfile> [<infile>]+\n", argv[0]);
		cleanup();
		exit(10);
	}
	infile_count = argc - 2;

	count_array = calloc(infile_count, sizeof (int));
	if (!count_array) {
		printf("Can't allocate count_array!\n");
		cleanup();
		exit(10);
	}

	memptr_array = calloc(infile_count, sizeof (void *));
	if (!memptr_array) {
		printf("Can't allocate memptr_array!\n");
		cleanup();
		exit(10);
	}

	fh_array = calloc(infile_count, sizeof (void *));
	if (!fh_array) {
		printf("Can't allocate fh_array!\n");
		cleanup();
		exit(10);
	}

	buf_array = malloc(memblocksize * (infile_count + 1));
	if (!buf_array) {
		printf("Can't allocate temporary buffers!\n");
		cleanup();
		exit(10);
	}

	/* do actual CD access */
	image_handle = fopen(argv[1], "ab");
	if (!image_handle) {
		printf("Can't open image file!\n");
		cleanup();
		exit(10);
	}

	/* now open file handles for all temporary files */
	for (j = 0; j < infile_count; j++) {
		fh_array[j] = fopen(argv[j + 2], "rb");
		if (!fh_array[j]) {
			printf("Can't open image file!\n");
			cleanup();
			exit(10);
		}
	}

	/* do actual majority vote algorithm */
	min_count_max = infile_count;
	while(1) {
		int size, count_max, loc_max;

		for (j = 0; j < infile_count; j++) {
			memptr_array [j] = buf_array + memblocksize * (j + 1);
			size = fread(memptr_array[j], 1, memblocksize, fh_array[j]);
		}
		if (size == 0) break;
		for (i = 0; i < size; i++) {
			for (j = 0; j < infile_count; j++)
				count_array[j] = 1;
			for (j = 0; j < infile_count; j++) {
				for (k = j + 1; k <infile_count; k++) {
					if (memptr_array[j][i] == memptr_array[k][i])
						count_array[j]++;
				}
				// have we achived the maximum possible?
				if (count_array[j] == infile_count - j)
					break;
			}
			count_max = 0;
			for (j = 0; j < infile_count; j++) {
				if (count_array[j] > count_max) {
					count_max = count_array[j];
					loc_max = j;
				}
			}
			buf_array[i] = memptr_array[loc_max][i];
			if (min_count_max > count_max)
				min_count_max = count_max;
		}
		size = fwrite(buf_array, 1, size, image_handle);
	}

	printf("min_count_max: %d\n", min_count_max);
	cleanup();
	exit(0);
}
